#  Copyright (c) 2020 Johannes Nolte
#  SPDX-License-Identifier: GPL-3.0-or-later

from mmusicc.formats._misc import init, MusicFile, types, loaders, mimes, AudioFileError

__all__ = ["init", "MusicFile", "types", "loaders", "mimes", "AudioFileError"]



#  Copyright (c) 2020 Johannes Nolte
#  SPDX-License-Identifier: GPL-3.0-or-later

import errno
import logging
import subprocess


class FFmpeg(object):
    """Wrapper for `FFmpeg <https://www.ffmpeg.org/>`.

    Args:
        source     (pathlib.path): path to source file
        target     (pathlib.path): path to target file
        options    (str): string containing options for ffmpeg as use in
            console (arguments separated by space)
        executable (str, optional): path to ffmpeg executable. Defaults to
            'ffmpeg': Can be overwritten in case e.g libav is used.
    """

    def __init__(self, source, target, options=None, executable="ffmpeg"):
        """Initialize ffmpeg command line wrapper.

        """
        self.executable = executable
        self._source = source
        self._target = target
        self._cmd = [executable]
        self.exit_status = -1

        self._cmd.extend(["-i", str(source)])
        if options:
            self._cmd.extend(options.split())
        self._cmd.append(str(target))

        self.cmd = subprocess.list2cmdline(self._cmd)
        self.process = None

    def __repr__(self):
        return "<{0!r} {1!r}>".format(self.__class__.__name__, self.cmd)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.exit_status == 0:
            if self._target.exists():
                self._target.unlink()

    def run(self):
        """Execute ffmpeg command line. Log stderr output.

        Raises:
            FFRuntimeError           : in case ffmpeg command exits with a
                non-zero code.
            FFExecutableNotFoundError: in case the executable path passed
                as not valid.
        """

        try:
            process_result = subprocess.run(self._cmd, stderr=subprocess.PIPE)
        except OSError as e:
            if e.errno == errno.ENOENT:
                raise FFExecutableNotFoundError(
                    "Executable '{0}' not found".format(self.executable)
                )
            else:
                raise

        logging.debug(process_result.stderr.decode())

        self.exit_status = process_result.returncode
        if self.exit_status != 0:
            raise FFRuntimeError(
                self.cmd,
                process_result.returncode,
                process_result.stdout,
                process_result.stderr,
            )


class FFExecutableNotFoundError(Exception):
    """Raise when ffmpeg executable was not found."""


class FFRuntimeError(Exception):
    """Raise when ffmpeg command line execution returns a non-zero exit code.

    """

    def __init__(self, cmd, exit_code, stdout, stderr):
        self.cmd = cmd
        self.exit_code = exit_code
        self.stdout = (stdout or b"").decode()
        self.stderr = (stderr or b"").decode()

        message = "`{0}` exited with status {1}\n\nSTDOUT:\n{2}\n\nSTDERR:\n{3}".format(
            self.cmd, exit_code, stdout, stderr
        )

        super(FFRuntimeError, self).__init__(message)



#  Copyright (c) 2020 Johannes Nolte
#  SPDX-License-Identifier: GPL-3.0-or-later

"""
recommended init order:
- init_logging
- init_formats
- init_allocationmap
"""


def init_allocationmap(path_config):
    """initialize allocation map of tags.

    Args:
        path_config (str): path to allocation map config yaml file
    """
    from mmusicc.util import allocationmap

    allocationmap.init_allocationmap(path_config)


def init_formats():
    """initialize formats module"""
    from mmusicc.formats import init

    init()


def init_logging(level=25, file_path=None):
    """initialize logger"""
    from mmusicc.util import init_logger

    return init_logger(level=level, file_path=file_path)



from mmusicc.database import MetaDB


test_dict = {
    "album": "mmusicc",  # str
    "artist": 42,  # int
    "title": 3.14,  # float
    "genre": ["fuu", "bar"],  # list
    "albumartist": ("fuu", "bar"),  # tuple
    "date": object(),  # object
    "albumartistsort": None,  # None
}


def test_dummy_load_allocation_map(allocation_map):
    # test not needed her but otherwise fixture (module level) will not load it
    assert len(allocation_map.list_tags) > 0


def test_read_and_write_database(temp_database):
    db = MetaDB(str(temp_database))
    db.insert_meta(test_dict, "some_path")
    db = MetaDB(str(temp_database))
    read_dict = db.read_meta("some_path")
    assert isinstance(read_dict.get("date"), object)
    read_dict.pop("date")
    for key in read_dict.keys():
        assert read_dict.get(key) == test_dict.get(key)
    assert read_dict.get("tracknumber") is None



#  Copyright (c) 2020 Johannes Nolte
#  SPDX-License-Identifier: GPL-3.0-or-later

import distutils.util
import logging
import sys

if __package__ is None and not hasattr(sys, "frozen"):
    # direct call of __main__.py
    import os.path

    path = os.path.realpath(os.path.abspath(__file__))
    sys.path.insert(0, os.path.dirname(os.path.dirname(path)))


def main(args=None):

    if args is None:
        args = sys.argv[1:]

    import mmusicc
    from mmusicc.util.logger import del_log_file, get_log_file

    try:
        m = mmusicc.MmusicC(args)
        if get_log_file():
            if not m.pre_result_logfile and m.error > 0:
                if distutils.util.strtobool(
                    input(
                        "One or more errors occurred during synchronisation.\n"
                        "Save log file to current working directory? [y/n] "
                    )
                ):
                    print(f"Log file can be found at: {get_log_file().resolve()}")
                else:
                    del_log_file()
            else:
                del_log_file()

        sys.exit(0)

    except Exception:
        logging.error("Exception occurred", exc_info=True)
        if "--log-file" in args:
            print("A error occurred in python, stack trace saved to log file.")
        else:
            if get_log_file():
                if distutils.util.strtobool(
                    input(
                        "A error occurred in python, save log file with stack trace to "
                        "current working directory? [y/n] "
                    )
                ):
                    print(f"Log file can be found at: {get_log_file().resolve()}")
                else:
                    del_log_file()

        sys.exit(1)


if __name__ == "__main__":
    main()



#  Copyright (c) 2020 Johannes Nolte
#  SPDX-License-Identifier: GPL-3.0-or-later

import argparse
import datetime
import enum
import logging
import math
import os
import pathlib
import textwrap

from mmusicc._init import init_formats, init_logging, init_allocationmap
from mmusicc.formats import AudioFileError
from mmusicc.formats import loaders as audio_loader
from mmusicc.formats import types as audio_types
from mmusicc.metadata import Metadata, AlbumMetadata
from mmusicc.util.allocationmap import get_tags_from_strs
from mmusicc.util.ffmpeg import FFmpeg, FFRuntimeError
from mmusicc.util.misc import is_supported_audio, swap_base, process_white_and_blacklist
from mmusicc.version import __version__ as package_version

str_description_rqw = textwrap.dedent(
    """\
    Metadata and file syncing the following combination are possible:
      - file   --> file
      - file   --> parent folder (target name is generated from source)
      - folder --> folder        (use --album to not to move through tree)
      - folder --> db            (full path as primary key)
      - db     --> folder        (key matching starts at leave of path)

    Supported Formats: 
    {}
    in Containers:
    {}
    """
)


# noinspection PyMethodMayBeStatic
class MmusicC:

    # TODO allow a predefined config file with all parser options to be passed

    def __init__(self, args):
        """Create a mmusicc object with the given options.
        Args:
            args (:list: string):
        """

        # the arguments the pre parser parses are identical to the main parser. This
        # allows for a nice help printout, but certain args parsed early on.
        pre_parser = argparse.ArgumentParser(add_help=False)
        pre_parser.add_argument(
            "-v", "--verbose", action="count", default=0,
        )
        pre_parser.add_argument(
            "--log-file", action="store",
        )
        parsed, remaining = pre_parser.parse_known_args(args)
        self.pre_result_verbose = parsed.verbose
        self.pre_result_logfile = parsed.log_file

        if self.pre_result_verbose == 0:
            log_level = 25
        elif self.pre_result_verbose == 1:
            log_level = logging.INFO
        else:
            log_level = logging.DEBUG

        self._log_path = init_logging(log_level, file_path=self.pre_result_logfile)
        init_formats()

        str_description = str_description_rqw.format(
            textwrap.fill(str(sorted([t.format for t in audio_types]))),
            textwrap.fill(str(sorted(list(audio_loader)))),
        )

        # noinspection PyTypeChecker
        self.parser = argparse.ArgumentParser(
            prog="mmusicc",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=str_description,
            # epilog="",
            add_help=False,
            allow_abbrev=True,
        )

        pg_required = self.parser.add_argument_group("Required Options")
        pg_general = self.parser.add_argument_group("General Options")
        pg_conversion = self.parser.add_argument_group("File Conversion")
        pg_meta = self.parser.add_argument_group("Metadata Syncing")

        group_source = pg_required.add_mutually_exclusive_group(required=True)
        group_source.add_argument(
            "-s", "--source", action="store", help="source file/album/lib-root.",
        )
        group_source.add_argument(
            "-sdb",
            "--source-db",
            action="store",
            help="source database (SQLite database file or (not tested) database URL.",
        )

        group_target = pg_required.add_mutually_exclusive_group(required=True)
        group_target.add_argument(
            "-t", "--target", action="store", help="target file/album/lib-root.",
        )
        group_target.add_argument(
            "-tdb", "--target-db", action="store", help="target database.",
        )

        pg_general.add_argument(
            "-h",
            "--help",
            action="help",
            default=argparse.SUPPRESS,
            help="Show this help message and exit.",
        )
        pg_general.add_argument("--version", action="version", version=package_version)
        pg_general.add_argument(
            "--album", action="store_true", help="only sync folder level",
        )
        group1 = pg_general.add_mutually_exclusive_group()
        group1.add_argument(
            "--only-meta",
            action="store_true",
            help="only sync meta, don't sync files. "
            "Auto set when syncing from/to database.",
        )
        group1.add_argument(
            "--only-files",
            action="store_true",
            help="only sync files, don't update metadata.",
        )
        pg_general.add_argument(
            "--dry-run",
            action="store_true",
            help="do everything as usual, but without writing (file and "
            "database). It is recommended to use with --only-meta or "
            "--only-files, otherwise errors are likely.",
        )
        pg_general.add_argument(
            "-v",
            "--verbose",
            action="count",
            default=0,
            help="print log messages. can be stacked up to 2 (info, debug).",
        )
        pg_general.add_argument(
            "-a", "--all", action="store_true", help="print log for unchanged files.",
        )
        pg_general.add_argument(
            "--log-file",
            action="store",
            help="Log file for detailed logging at DEBUG level. "
            "If file exist log is appended.",
        )

        pg_conversion.add_argument(
            "-f",
            "--format",
            action="store",
            required=False,
            help="output container format of ffmpeg conversion "
            "(ignored when target is file_path).",
        )
        pg_conversion.add_argument(
            "-o",
            "--ffmpeg-options",
            action="store",
            help="conversion options for ffmpeg conversion, if empty ffmpeg defaults "
            "are used. It is recommended to test them directly with ffmpeg before "
            "they are used with mmusicc.",
        )

        tmp_double_txt_2 = (
            "Can be passed as file (Plain text file, "
            "#containing "
            "tags separated with a new line) or as one or "
            "multiple arguments."
        )
        pg_meta.add_argument(
            "--white-list-tags",
            nargs="+",
            action="store",
            help="Tags to be whitelisted. " + tmp_double_txt_2,
        )
        pg_meta.add_argument(
            "--black-list-tags",
            nargs="+",
            action="store",
            help="Tags to be blacklisted. " + tmp_double_txt_2,
        )
        pg_meta.add_argument(
            "--lazy-import",
            action="store_true",
            help="do not overwrite an existing non-None tag with None when importing "
            "metadata. Therefore,  if a tag is not excluded in the source it is "
            "not deleted in the target and the tag in the target remains "
            "unchanged.",
        )
        pg_meta.add_argument(
            "--delete-existing-metadata",
            action="store_true",
            help="delete existing metadata from target. This includes all defined, "
            "and unprocessed tags.",
        )
        pg_meta.add_argument(
            "--path-config", action="store", help="file path to custom config file.",
        )

        logging.debug("mmusicc running with arguments: {}".format(args))

        self.result = self.parser.parse_args(args)

        Metadata.dry_run = self.result.dry_run

        if not self.result.path_config:
            path_folder_this_file = pathlib.Path(__file__).resolve().parent
            self.result.path_config = path_folder_this_file.joinpath(
                "data", "config.yaml"
            )

        try:
            init_allocationmap(self.result.path_config)
        except FileNotFoundError:
            self.parser.error("path to config file not found")

        self.run_files = not self.result.only_meta
        self.run_meta = not self.result.only_files

        self.db_url = None
        if self.result.source:
            self.source = pathlib.Path(self.result.source).expanduser().resolve()
        else:
            self.source = None
            self.db_url = self.result.source_db

        if self.result.target:
            self.target = pathlib.Path(self.result.target).expanduser().resolve()
        else:
            self.target = None
            self.db_url = self.result.target_db

        if self.db_url:
            if Metadata.is_linked_database:
                Metadata.unlink_database()
            Metadata.link_database(self.db_url)

        self.source_type = self.get_element_type(self.source)
        self.target_type = self.get_element_type(self.target)

        if not (self.source or self.target) and self.run_files:
            self.parser.error("Target or source is database! I can only run meta!")

        if not self.source and not self.target:
            self.parser.error("Can't sync from database to database!")

        if self.source and self.target:
            self.format_extension = None
            if self.target_type == MmusicC.ElementType.file:
                if "." in self.target.suffix:  # target is file
                    self.format_extension = self.target.suffix

            if not self.format_extension and self.result.format:
                # target is folder an format has to be given!
                self.format_extension = self.result.format
                if "." not in self.format_extension:
                    self.format_extension = "." + self.format_extension

            if not self.format_extension:
                self.parser.error(
                    "the following arguments are required: "
                    "-f/--format, except at file-->file and "
                    "album/lib-->database operations."
                )

        self.whitelist = None
        try:
            if self.result.white_list_tags:
                self.whitelist = load_tags_from_list_or_file(
                    self.result.white_list_tags
                )
        except KeyError as err:
            self.parser.error(
                "Whitelist: {}. If input is file it is not found.".format(err)
            )

        self.blacklist = None
        try:
            if self.result.black_list_tags:
                self.blacklist = load_tags_from_list_or_file(
                    self.result.black_list_tags
                )
        except KeyError as err:
            self.parser.error(
                "Blacklist: {}. If input is file it is not found.".format(err)
            )

        # just fot the log
        if self.whitelist or self.blacklist:
            whitelist = process_white_and_blacklist(self.whitelist, self.blacklist)
            logging.info("Tags to be Synced: {}".format(whitelist))

        # stats for report
        self.unchanged = 0
        self.created = 0
        self.metadata = 0
        self.both = 0
        self.error = 0
        time_start = datetime.datetime.now()

        string_opt_args = ""
        for att in [
            "album",
            "delete_existing_metadata",
            "dry_run",
            "lazy_import",
            "only_files",
            "only_meta",
            "log_file",
            "all",
        ]:
            if getattr(self.result, att, None):
                string_opt_args += f"{att}={getattr(self.result, att)}; "

        options = [
            f"             Running mmusicc",
            f"source type: {self.source_type}",
            f"source path: {self.source if self.source else self.db_url}",
            f"target type: {self.target_type}",
            f"target path: {self.target if self.target else self.db_url}",
            f"format     : {getattr(self, 'format_extension', '')}"
            f"{' | ' if self.result.ffmpeg_options else ''}"
            f"{self.result.ffmpeg_options if self.result.ffmpeg_options else ''}",
            f"options    : {string_opt_args[:-2]}",
            f"path config: {self.result.path_config}",
        ]

        for o in options:
            logging.log(25, o)

        logging.log(25, "---------------------------------------------------------")

        if self.db_url:
            if self.source_type == MmusicC.ElementType.file:
                self.handle_media2db(self.source)
            elif self.source_type == MmusicC.ElementType.folder:
                if self.result.album:
                    self.handle_media2db(self.source)
                else:
                    gen = os.walk(self.source, topdown=True)
                    for root, dirs, files in gen:
                        audio_files = [
                            file for file in files if is_supported_audio(file)
                        ]
                        if len(audio_files) > 0:
                            self.handle_media2db(root)
            elif self.target_type == MmusicC.ElementType.file:
                self.handle_db2media(self.target)
            elif self.target_type == MmusicC.ElementType.folder:
                if self.result.album:
                    self.handle_db2media(self.target)
                else:
                    gen = os.walk(self.target, topdown=True)
                    for root, dirs, files in gen:
                        audio_files = [
                            file for file in files if is_supported_audio(file)
                        ]
                        if len(audio_files) > 0:
                            self.handle_db2media(root)

        elif self.source_type == MmusicC.ElementType.file:
            self.handle_files2file(self.source, self.target)

        elif self.source_type == MmusicC.ElementType.folder:
            if self.result.album:
                self.handle_album2album(self.source, self.target)
            else:
                logging.debug(
                    "Walking through tree of directory '{}'".format(self.source)
                )
                gen = os.walk(self.source, topdown=True)
                for root, dirs, files in gen:
                    root = pathlib.Path(root)
                    album_target = swap_base(self.source, root, self.target)
                    logging.info("Current root: {}".format(root))

                    if any(is_supported_audio(f) for f in files):
                        if self.target_type == MmusicC.ElementType.folder:
                            self.handle_album2album(root, album_target)
                        else:
                            self.handle_media2db(root)

                    # i like when singles have their own folder like albums have
                    if len(files) > 0 and len(dirs) > 0:
                        logging.info(
                            "Found files not in album {} at folder '{}'".format(
                                files, root
                            )
                        )

        time_delta = datetime.datetime.now() - time_start

        if self.created + self.metadata + self.both > 0 or self.result.all:
            logging.log(25, "---------------------------------------------------------")

        report = [
            f"Total Time : {math.floor(time_delta.total_seconds() / 60)} min "
            f"{math.fmod(time_delta.total_seconds(), 60)} s",
            f"Unchanged  : {self.unchanged}",
            f"Metadata   : {self.metadata}",
            f"Created    : {self.created}",
            f"Both       : {self.both}",
            f"Errors     : {self.error}",
        ]

        if self.target is MmusicC.ElementType.database:
            logging.log(25, report[1])
        else:
            for r in report:
                logging.log(25, r)

    # __init___

    def handle_files2file(self, file_source, file_target):
        if not is_supported_audio(file_source):
            logging.warning(f"file '{file_source}' not a supported audio file.")
            return -1
        logging.debug(f"handle_files2file: {file_source}->{file_target}")
        # compute filename when only target folder is given
        if self.get_element_type(file_target) == MmusicC.ElementType.folder:
            file_target = file_target.joinpath(file_source.stem + self.format_extension)
        change = 0
        change += self._handle_files2file_file(file_source, file_target)
        if change < 0:
            return change
        change += self._handle_files2file_meta(file_source, file_target)
        self.log_changes(change, file_target, make_relative=False)
        return change

    def _handle_files2file_file(self, file_source, file_target):
        if self.run_files and file_source and file_target:
            if file_target.is_file():
                logging.debug(f"ffmpeg skipped target file exists: '{file_target}'")
            elif not self.result.dry_run:
                with FFmpeg(
                    file_source, file_target, options=self.result.ffmpeg_options,
                ) as ffmpeg:
                    try:
                        ffmpeg.run()
                        return 2
                    except FFRuntimeError as ex:
                        tmp_msg = [str(file_target.relative_to(self.target))]
                        last3 = ex.stderr.split("\n")[-4:-1]
                        tmp_msg.extend(last3)
                        tmp_msg_str = ("\n" + " " * 5).join(tmp_msg)
                        logging.log(25, f"ffmpeg error: {tmp_msg_str}")
                        return -2
        return 0

    def _handle_files2file_meta(self, file_source, file_target):
        if self.run_meta:
            try:
                meta_source = Metadata(file_source)
                meta_target = Metadata(file_target)
                meta_target.import_tags(
                    meta_source,
                    whitelist=self.whitelist,
                    blacklist=self.blacklist,
                    skip_none=self.result.lazy_import,
                    clear_blacklisted=self.result.delete_existing_metadata,
                )
                return meta_target.write_tags(
                    remove_existing=self.result.delete_existing_metadata
                )
            except AudioFileError as ex:
                logging.info(ex)
                return -1
            except FileNotFoundError as ex:
                logging.info(ex)
                return -4
        return 0

    def handle_album2album(self, album_source, album_target):
        """
        Note:
            For the moment, it makes no difference if metadata is synced file by file as
            Metadata or album wise as AlbumMetadata. This will change, when the
            interactive mode with album wise comparison is introduced.
            Therefore the metadata syncing is currently done in run_files, except when
            the only-meta option is given.
        """
        logging.debug(f"handle_album2album: {album_source}->{album_target}")
        if self.run_files and album_source and album_target:
            changes_file = dict()
            if not album_target.is_dir():
                album_target.mkdir(parents=True)
            for file in sorted(os.listdir(album_source)):
                if is_supported_audio(file):
                    file_source = album_source.joinpath(file)
                    file_target = album_target.joinpath(
                        file_source.stem + self.format_extension
                    )
                    changes_file[file_target] = self._handle_files2file_file(
                        file_source, file_target
                    )
                    if self.run_meta:
                        changes_file[file_target] += self._handle_files2file_meta(
                            file_source, file_target
                        )
                    self.log_changes(changes_file[file_target], file_target)
                else:
                    logging.info(f"File is not supported or not valid: {file}")

        if self.run_meta and not self.run_files:
            if not album_target.is_dir():
                logging.warning(
                    "no target folder for given source '{}', "
                    "skipping".format(album_source)
                )
                return
            meta_source = AlbumMetadata(album_source)
            meta_target = AlbumMetadata(album_target)
            meta_target.import_tags(
                meta_source,
                whitelist=self.whitelist,
                blacklist=self.blacklist,
                skip_none=self.result.lazy_import,
                clear_blacklisted=self.result.delete_existing_metadata,
            )
            changes_meta = meta_target.write_tags(
                remove_existing=self.result.delete_existing_metadata
            )
            for file, change in changes_meta.items():
                self.log_changes(change, file)

    def handle_media2db(self, album_source):
        album_target = pathlib.Path(album_source)
        if album_target.is_file():
            meta_source = Metadata(album_source)
        else:
            meta_source = AlbumMetadata(album_source)
        meta_source.export_tags_to_db()

    def handle_db2media(self, album_target):
        album_target = pathlib.Path(album_target)
        if album_target.is_file():
            meta_target = Metadata(album_target)
        else:
            meta_target = AlbumMetadata(album_target)
        meta_target.import_tags_from_db(
            whitelist=self.whitelist,
            blacklist=self.blacklist,
            skip_none=self.result.lazy_import,
            clear_blacklisted=self.result.delete_existing_metadata,
        )
        meta_target.write_tags(remove_existing=self.result.delete_existing_metadata)

    def log_changes(self, change, file, make_relative=True):
        if file.is_absolute() and make_relative:
            file = file.relative_to(self.target)
        if change < 0:
            self.error += 1
        elif change > 0:
            if change == 1:
                self.metadata += 1
            elif change == 2:
                self.created += 1
            elif change == 3:
                self.both += 1
        else:
            self.unchanged += 1
            if not self.result.all:
                logging.info(f"{change} > {file}")
                return

        logging.log(25, "{: d} > {}".format(change, file))

    def get_element_type(self, element):
        """Get the element type based on the pathlib.Path object.

        Instance method in clarifying that the function is instance dependent.

        """

        if not element:
            if Metadata.is_linked_database:
                return MmusicC.ElementType.database
            else:
                return MmusicC.ElementType.other

        if element.suffix in audio_loader.keys():  # target is file
            if element.exists():
                if not is_supported_audio(element):
                    self.parser.error(f"File '{element}' not supported!")
            return MmusicC.ElementType.file
        else:
            if "." in element.suffix:
                logging.info(
                    f"Path identified as folder, although it can also be an "
                    f"unsupported file with extension {element.suffix}"
                )
            return MmusicC.ElementType.folder

    class ElementType(enum.Enum):
        file = (1,)
        folder = (2,)
        database = (3,)
        other = 4


def load_tags_from_list_or_file(list_tags_or_file):
    """Load TAG-list from a list of STRING or a file.

    Where each tag (STRING) is separated by newline.

    Args:
        list_tags_or_file (list of str or str): list of tags or file_path

    Returns:
        list of str: TAGs

    """
    if len(list_tags_or_file) == 1:
        path = pathlib.Path(list_tags_or_file[0])
        if path.exists():
            with path.open(mode="r") as f:
                list_tags_or_file = [line.rstrip() for line in f]
    return get_tags_from_strs(list_tags_or_file)



import logging
import pathlib

from sqlalchemy import MetaData, Table, Column, String, PickleType
from sqlalchemy import create_engine

from mmusicc.util.allocationmap import list_tags


class MetaDB:
    """Object representing a database connection in Metadata,

    holds connection parameters and inserts and reads data. When writing to the
    Database, always all data is written, while you can load only select which
    tags to load. See https://docs.sqlalchemy.org/en/13/core/engines.html for
    Database Url examples.

    There will be two tables. One only containing strings and string
    representation of objects and one containing the pickled strings or
    objects. This way all object types can be restored at loading while at the
    same time the database can be search with standard database queries.

    Args:
        database_url (str): database url following RFC-1738*. If the sting,
            does not contain '://', a filepath for a sqlite database is
            assumed.
    """

    def __init__(self, database_url):
        if "://" not in database_url:
            # it is assumed the database url is filepath and therefore SQLite
            database_url = pathlib.Path(database_url).expanduser().resolve()
            database_url = "sqlite:///" + str(database_url)
        self._database_url = database_url
        self._engine = create_engine(self._database_url)
        if not list_tags:
            logging.warning("no tags found! Is project initialized")
        self._create_table(list_tags)

    @property
    def database_url(self):
        """Get URL of connected database."""
        return self._engine.url

    def _create_table(self, list_keys):
        """Create a Tables in Database if it does not already exists,

        where each column represents a tag (=key in list_keys).

        Args:
            list keys (list of str): list of tags in Metadata
        """
        self.list_keys = list_keys
        sql_metadata = MetaData()
        self.tags = Table(
            "tags", sql_metadata, Column("_primary_key", String(200), primary_key=True)
        )
        self.pickle_tags = Table(
            "pickle_tags",
            sql_metadata,
            Column("_primary_key", String(200), primary_key=True),
        )
        for key in list_keys:
            self.tags.append_column(Column(key, String(100)))
            self.pickle_tags.append_column(Column(key, PickleType()))

        self.tags.create(self._engine, checkfirst=True)
        self.pickle_tags.create(self._engine, checkfirst=True)

    def insert_meta(self, dict_data, primary_key):
        """Inserts a row into the database, with the values from the dict.

        Args:
            dict_data  (dict): metadata dictionary (`Dict[str, object]`) to be
                written.
            primary_key (str): unique identifier of the item which data is to
                be written. Metadata uses the absolute filepath.
        """
        dict_meta_pickle = dict_data.copy()
        dict_meta_pickle["_primary_key"] = primary_key
        dict_meta = dict()
        dict_meta["_primary_key"] = primary_key
        for key in list(dict_meta_pickle):
            dict_meta[key] = str(dict_meta_pickle[key])

        with self._engine.connect() as conn:
            conn.execute(self.tags.insert().values(dict_meta))
            conn.execute(self.pickle_tags.insert().values(dict_meta_pickle))

    def read_meta(self, primary_key, tags=None):
        """returns values of a row with given primary key as metadata dict.

        Args:
            primary_key         (str): unique identifier of the item which data
                is to read (eg. Filepath).
            tags        (list of str): list of strings to be read, reads all if
                None. Defaults to None.

        Returns:
            dict_data (dict<str:obj>): metadata dictionary
        """
        with self._engine.connect() as conn:
            foo_col = Column("_primary_key")
            result = conn.execute(
                self.pickle_tags.select().where(foo_col == primary_key)
            ).first()
            if result:
                dict_data_tmp = dict(result)
                dict_data_tmp.pop("_primary_key")
                if tags:
                    for key in list(dict_data_tmp):
                        if key not in tags:
                            dict_data_tmp.pop(key)
                return dict_data_tmp
            return None

    def get_list_of_primary_keys(self):
        """reads the primary keys from database and returns them.

        Returns:
            list of str: list of primary key strings in database
        """
        with self._engine.connect() as conn:
            result = list(conn.execute("SELECT _primary_key FROM tags"))
            return [s[0] for s in result]



#  Copyright (c) 2020 Johannes Nolte
#  SPDX-License-Identifier: GPL-3.0-or-later

import re
import warnings
from distutils.file_util import copy_file

import mutagen
import pytest

import mmusicc.formats
import mmusicc.util.allocationmap
from mmusicc.metadata import Empty
from mmusicc.util.metadatadict import metadatadict
from ._util import *


def test_dummy_for_init(allocation_map, audio_loaders):
    """test not needed her but otherwise fixture (module level) will not be
        loaded and the init functions are not called
    """
    assert len(allocation_map.list_tags) == 23
    assert len(audio_loaders) == 9


@pytest.fixture(scope="class")
def audio_files(audio_loaders, dir_subpackages) -> dict:
    """collect all format test files from subpackages folder"""
    loaders = set(audio_loaders.values())
    extensions = set(audio_loaders.keys())
    files = [p.name for p in dir_subpackages.glob("*.*")]
    regex = re.compile(r"formats_[\w\d]*.[\w\d]{3,4}")
    files = list(filter(regex.search, files))
    audio_files = dict()
    for f in files:
        path = dir_subpackages.joinpath(f)
        audio_files[path.suffix] = path
        try:
            extensions.remove(path.suffix)
            loaders.remove(audio_loaders.get(path.suffix))
        except KeyError:
            pass
    # if loaders is empty all loaders have/will been tested
    # extensions is just for info what supported extension types are not tested
    if len(loaders) > 0:
        warnings.warn(
            UserWarning("module(s) '{}' is not been tested".format(str(loaders)))
        )
    return audio_files


@pytest.fixture(scope="module")
def expected_metadata(dir_orig_data) -> dict:
    """expected results for a subset of the loaded elements"""
    data = {
        "album": "str_album",
        "albumartist": "str_albumartist",
        "albumartistsort": "str_albumartistsort",
        "artist": "str_artist",
        "bpm": "128",
        "comment": "str_comment",
        "composer": ["str_composer_A", "str_composer_B"],
        "date": "2020",
        "discid": "str_discid",
        "discnumber": "2",
        "genre": "str_genre",
        "isrc": "QZES81947811",
        "lyrics": "str_lyrics",
        "title": "str_title",
        "tracknumber": "3",
    }
    return data


@pytest.fixture(scope="module")
def metadata_write_tags(expected_metadata) -> dict:
    """adds _2 to strings and increases numbers by one and sets artist None"""
    _dict = expected_metadata.copy()
    for key in list(_dict):
        if isinstance(_dict[key], list):
            continue
        try:
            cur_int = int(_dict[key])
            _dict[key] = str(cur_int + 1)
        except ValueError:
            if key == "originaldate":
                continue
            _dict[key] = _dict[key] + "_2"
    _dict["artist"] = Empty()
    _dict["albumartist"] = None
    return _dict


@pytest.fixture(scope="function")
def media_file(request, audio_files, dir_lib_test):
    """find test file with the given extension and copy it to a test folder"""
    file = audio_files.get(request.param)
    if not file:
        pytest.xfail("No file with given extension '{}' exists.".format(request.param))
    copy_file(str(file), str(dir_lib_test))
    return dir_lib_test.joinpath(file.name)


@pytest.fixture(scope="function")
def media_file_th(media_file):
    return save_files_hash_and_mtime([media_file])


# TODO find a way to load extension dynamically
# TODO run tests per-class configuration


@pytest.mark.parametrize(
    "media_file", [".mp3", ".flac", ".ogg", ".opus"], indirect=["media_file"]
)
class TestFormats:
    def test_read(self, media_file, media_file_th, expected_metadata):
        """read from file and compare with expected"""
        m_file = _assert_read_and_compare_file(media_file, expected_metadata)
        keys = list(m_file.unprocessed_tag)
        assert len(keys) == 1
        assert m_file.unprocessed_tag.get(keys[0]) == "not in tag list"
        assert cmp_files_hash_and_time(media_file, media_file_th) == 1

    def test_write_identical(self, media_file, media_file_th):
        """write read data unchanged, files should not be modified."""
        m_file = mmusicc.formats.MusicFile(media_file)
        m_file.file_read()
        m_file.file_save(remove_existing=False, write_empty=False)
        assert cmp_files_hash_and_time(media_file, media_file_th) == 1

    def test_write_identical_del_existing(self, media_file, media_file_th):
        """write read data unchanged, bur remove_existing forcing a write to the
        file. Files have to be modified but the content must be identical."""
        m_file_1 = mmusicc.formats.MusicFile(media_file)
        m_file_1.file_read()
        m_file_1.file_save(remove_existing=True, write_empty=False)
        assert cmp_files_hash_and_time(media_file, media_file_th) == 10101
        m_file_2 = mmusicc.formats.MusicFile(media_file)
        m_file_2.file_read()
        assert not set(m_file_1.dict_meta).difference(set(m_file_2.dict_meta))
        media_file_2_th = save_files_hash_and_mtime(media_file)
        m_file_2.file_save(remove_existing=True, write_empty=False)
        assert cmp_files_hash_and_time(media_file, media_file_2_th) == 1

    @pytest.mark.parametrize("remove_existing", [False, True])
    @pytest.mark.parametrize("write_empty", [False, True])
    def test_write(
        self,
        media_file,
        metadata_write_tags,
        remove_existing,
        write_empty,
        media_file_th,
    ):
        """write to file and compare with expected, test optional arguments"""
        _write_meta_to_file(
            media_file, metadata_write_tags, remove_existing, write_empty
        )
        # from test_read we already know that we reading works
        m_file = _assert_read_and_compare_file(
            media_file, metadata_write_tags, exclude=["artist", "albumartist"]
        )

        # in the test file is one tag placed (encoder settings) that is not in
        # the tag dictionary. This tag can be only removed with the
        # remove-existing option. Also with remove existing, a reread with the
        # same tag list will always lead 0 unprocessed_tag.

        assert m_file.dict_meta["album"] == "str_album_2"

        if remove_existing:
            assert len(list(m_file.unprocessed_tag)) == 0
            assert m_file.dict_meta["albumartist"] is None
        else:
            assert len(list(m_file.unprocessed_tag)) == 1
            assert m_file.dict_meta["albumartist"] == "str_albumartist"

        if media_file.suffix == ".mp3":
            # mp3 can not be tested for write empty
            return

        if write_empty:
            assert Empty.is_empty(m_file.dict_meta["artist"])
        else:
            assert m_file.dict_meta["artist"] is None

        assert cmp_files_hash_and_time(media_file, media_file_th) == 10101

    def test_read_and_write_no_header(self, media_file, expected_metadata):
        """try reading file with no header, header is deleted by mutagen"""
        file_type = mutagen.File(media_file)
        file_type.delete()
        file_type.save()
        _assert_read_and_compare_file(media_file, {})
        _write_meta_to_file(media_file, expected_metadata, True)
        _assert_read_and_compare_file(media_file, expected_metadata)

    def test_write_multiple_tag_values(self, media_file):
        """try writing multiple tag values (lists in tag dict),
        reading was tested in test_read()
        """
        val_dict = {"artist": ["fuu", "bar"]}
        _write_meta_to_file(media_file, val_dict, True)
        _assert_read_and_compare_file(media_file, val_dict)


def _write_meta_to_file(path, dict_meta, remove_existing, write_empty=True):
    """helper creates MusicFile object"""
    m_file = mmusicc.formats.MusicFile(path)
    m_file.dict_meta = metadatadict(dict_meta)
    m_file.file_save(remove_existing=remove_existing, write_empty=write_empty)


def _assert_read_and_compare_file(path, dict_answer, exclude=None):
    """helper reads a audio file and compare its contents with expected values
        given in dict_answer. Single tags can be excluded.
    """
    m_file = mmusicc.formats.MusicFile(path)
    m_file.file_read()
    if not exclude:
        exclude = []
    for tag in list(dict_answer):
        if tag in exclude:
            continue
        assert m_file.dict_meta.get(tag) == dict_answer.get(tag)
    return m_file



#  Copyright (c) 2020 Johannes Nolte
#  SPDX-License-Identifier: GPL-3.0-or-later

import datetime
import logging
import logging.config
import pathlib
import shutil


def copy_log_file(path_target):
    """copy current log file of FileHandler to target path

    Returns:
        pathlib.Path: actual path of target
    """
    org_log_path = getattr(logging.Logger, "manager").root.handlers[1].baseFilename
    path_target = pathlib.Path(path_target)
    if path_target.exists():
        path_target = path_target.parent.joinpath(
            path_target.stem
            + datetime.datetime.now().strftime("-%Y%m%d-%H%M%S")
            + path_target.suffix
        )
    shutil.copyfile(org_log_path, path_target)
    return path_target


def del_log_file():
    """delete current log file of FileHandler"""
    if get_log_file():
        get_log_file().unlink()


def get_log_file():
    """
    Returns:
        pathlib.Path: path of file logged to by FileHandler
    """
    manager = getattr(logging.Logger, "manager")

    if not manager.root.handlers:
        return None

    if "LogCaptureHandler" in str(manager.root.handlers[0]):
        # mmusicc was started from pytest. pytest adds a LogCaptureHandler to logger.
        # Which causes init logger to think the logger is already initialized and does
        # not add root or file handlers. Since its not wanted when testing anyways,
        # this is quite convenient. As there is no file handler there is no file.
        return None

    return pathlib.Path(
        getattr(logging.Logger, "manager").root.handlers[1].baseFilename
    )


def init_logger(level=25, file_path=None):
    """Initialise Python Logger (function may be extended in future)

    Args:
        level (int): log level for console output
        file_path(str or pathlib.Path or None): file path of log file. If None a
            logfile with a unique name is created at the current working directory,
            which must be deleted manually. If path exits log is append to file.
            Defaults to None.

    Returns:
        pathlib.Path: path of log file
    """

    if len(logging.getLogger().handlers):
        logging.debug("Logger Already Initialized, Skipping.")
        return

    if not file_path:
        file_path = pathlib.Path(
            "mmusicc" + datetime.datetime.now().strftime("-%Y%m%d-%H%M%S") + ".log"
        )

    file_path = pathlib.Path(file_path)
    file_path = file_path.expanduser().resolve()

    # noinspection PyPep8
    config = {
        "version": 1,
        "formatters": {
            "simple": {
                "format": "%(asctime)s.%(msecs)03d %(levelname)-8s %(module)-13s %(message)s",
                "datefmt": "%H:%M:%S",
            },
            "short": {
                "format": "%(asctime)s.%(msecs)03d %(message)s",
                "datefmt": "%H:%M:%S",
            },
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "level": level,
                "formatter": "short" if level == 25 else "simple",
                "stream": "ext://sys.stdout",
            },
            "file_handler": {
                "class": "logging.FileHandler",
                "filename": str(file_path),
                "formatter": "simple",
                "level": "DEBUG",
                "mode": "a",
            },
        },
        "root": {"level": "DEBUG", "handlers": ["console", "file_handler"],},
    }
    logging.config.dictConfig(config)
    logging.debug("Initialized logger")
    logging.addLevelName(25, "mmusicc")

    return file_path



import os
import sys

from mmusicc.util import importhelper
from mmusicc.util.allocationmap import init_allocationmap
from mmusicc.util.logger import init_logger


# noinspection PyProtectedMember
def get_module_dir(module=None):
    """Returns the absolute path of a module. If no module is given
    the one this is called from is used.
    """

    if module is None:
        file_path = sys._getframe(1).f_globals["__file__"]
    else:
        file_path = getattr(module, "__file__")
    return os.path.dirname(os.path.realpath(file_path))



#  Copyright (c) 2020 Johannes Nolte
#  SPDX-License-Identifier: GPL-3.0-or-later

from ._init import init_allocationmap, init_formats, init_logging
from .metadata import Metadata
from .mmusicc import MmusicC

name = "mmusicc"



"""
Module adapted and copied from:
https://github.com/quodlibet/quodlibet/blob/master/quodlibet/util/importhelper.py

"""

import importlib.util as import_util
import logging
import os
import sys


def load_dir_modules(path, package):
    """Load all modules and packages in path (recursive).

    In case the module is already loaded, doesn't reload it.
    """

    try:
        modules = [e[0] for e in get_importables(path)]
    except OSError:
        logging.error("%r not found" % path)
        return []

    # get_importables can yield py and pyc for the same module
    # and we want to load it only once
    modules = set(modules)

    loaded = []
    for name in modules:
        try:
            mod = load_module(name, package, path)
        except Exception as ex:
            logging.error(ex)
            continue
        if mod:
            loaded.append(mod)

    return loaded


def get_importables(folder):
    """Searches a folder and its subfolders for modules and packages to import.
    No subfolders in packages, .so supported.

    The root folder will not be considered a package.

    returns a tuple of the name, import path, list of possible dependencies
    """

    def is_ok(f):
        if f.startswith("_"):
            return False
        if f.endswith(".py"):
            return True
        return False

    def is_init(f):
        if f == "__init__.py":
            return True
        return False

    first = True
    for root, dirs, names in os.walk(folder):
        # Ignore packages like "_shared"
        if os.path.basename(root).startswith("_"):
            continue
        if not first and any((is_init(n) for n in names)):
            yield (
                os.path.basename(root),
                root,
                list(filter(is_ok, [os.path.join(root, name) for name in names])),
            )
        else:
            for name in filter(is_ok, names):
                yield (
                    os.path.splitext(name)[0],
                    os.path.join(root, name),
                    [os.path.join(root, name)],
                )
        first = False


def load_module(name, package, path):
    """Load a module/package. Returns the module or None.
       Doesn't catch any exceptions during the actual import.
    """

    fullname = package + "." + name
    try:
        return sys.modules[fullname]
    except KeyError:
        pass

    module_spec = import_util.find_spec(fullname, [path])
    if module_spec is None:
        return

    module = import_util.module_from_spec(module_spec)

    # modules need a parent package
    if package not in sys.modules:
        sys.modules[package] = module

    module_spec.loader.exec_module(module)

    return module



#  Copyright (c) 2020 Johannes Nolte
#  SPDX-License-Identifier: GPL-3.0-or-later

import copy
import logging

import mutagen

import mmusicc.util.allocationmap as am
from mmusicc.formats._audio import AudioFile
from mmusicc.util.metadatadict import Empty, scan_dictionary, AlbumArt
from mmusicc.util.util import text_parser_get, join_str_list

extensions = [".mp3", ".mp2", ".mp1", ".mpg", ".mpeg"]
"""list of all extensions associated with this module"""
# loader   see bottom
# types    see bottom


def _to_album_art(frame):
    album_art = AlbumArt()
    album_art.desc = frame.desc
    album_art.ptype = frame.type
    album_art.data = frame.data
    album_art.mime = frame.mime
    return album_art


class MP3File(AudioFile):
    """Tag object for mp3 files using ID3 tag standard.

    Instance created by loader (at the bottom of this code) which in this case
    is this class itself (vgl. xiph).

    Args:
        file_path      (str): file path of represented audio file
    """

    format = "MP3"  # "or MPEG-1/3"
    mimes = ["audio/mp3", "audio/x-mp3", "audio/mpeg", "audio/mpg", "audio/x-mpeg"]

    def __init__(self, file_path):
        super().__init__()
        self._file = mutagen.File(file_path)
        self._changed_tags = None

    def file_read(self):
        """reads file tags into AudioFile tag dictionary (dict_meta).

        First tries to associate ID3 tags, than takes all txxx tags and runs
        them through the scan dictionary function.
        """
        tags_txxx = dict()

        for frame in self._file.values():

            frame_type = type(frame)
            frame_type_parent = frame_type.mro()[1]

            # handle each frame depending on parent class
            if frame.FrameID == "APIC":
                tag_val = _to_album_art(frame)
            elif frame_type is mutagen.id3.TXXX:
                tags_txxx[frame.desc] = frame.text[0]
                continue
            elif frame_type_parent is mutagen.id3.TextFrame:
                tag_val = frame.text[0]
            elif frame_type_parent is mutagen.id3.PairedTextFrame:
                tmp_key = "PairedTextFrame not supported:" + frame.HashKey
                self.unprocessed_tag[tmp_key] = frame.people
                continue
            elif frame_type_parent is mutagen.id3.NumericPartTextFrame:
                tag_val = frame.text[0]
            elif frame_type_parent is mutagen.id3.NumericTextFrame:
                tag_val = frame.text[0]
            elif frame_type_parent is mutagen.id3.TimeStampTextFrame:
                tag_val = frame.text[0].get_text()
            elif frame_type_parent is mutagen.id3.Frame:
                tag_val = frame.text
            else:
                raise Exception("frame not implemented %s", frame.HashKey)

            if Empty.is_empty(tag_val):
                tag_val = Empty()

            tag_key = am.dict_id32tag.get(frame.FrameID)
            if tag_key:
                if isinstance(tag_val, Empty) or isinstance(tag_val, AlbumArt):
                    val = tag_val
                else:
                    val = text_parser_get(tag_val)
                    if len(val) == 1:
                        val = val[0]
                self.dict_meta[tag_key] = val
            else:
                self.unprocessed_tag[frame.HashKey] = tag_val

        if len(tags_txxx) > 0:
            unprocessed_txxx = scan_dictionary(tags_txxx, self.dict_meta)
            # restore original tag hash to be able to delete it
            restored_txxx = dict()
            for key in tags_txxx:
                if key.casefold() in unprocessed_txxx:
                    restored_txxx["TXXX:" + key] = tags_txxx[key]
            self.unprocessed_tag.update(restored_txxx)

        return self

    def file_save(
        self, remove_existing=False, write_empty=False, remove_v1=False, dry_run=False
    ):
        """saves file tags from tag dictionary (dict_meta) to AudioFile.

        Note:
            write_empty may have no effect. Since mutagen will not load empty
            tags it can't be checked. Correction or Info is appreciated.

        Args:
            remove_existing ('bool', optional): if true clear all tags on file
                before writing. Defaults to False.
            write_empty     (bool): Only affects TXXX tags. Existing tags will
                always be set to None. If true create empty TXXX tags with
                value none. If false no tag will be created or existing
                TXXX tag on file will be deleted. Defaults to False.
            remove_v1       ('bool'): If True, remove existing ID3.V1 tags.
                Defaults to False.
            dry_run (bool): if true, do anything but saving to file. Defaults to False

        Returns:
            int: 1 if data was saved to file, zero if nothing was changed on file.
        """
        self._changed_tags = list()

        if not dry_run:
            audio = self._file
        else:
            audio = copy.deepcopy(self._file)

        if audio.tags is None:
            audio.add_tags()

        if audio.tags is None:
            audio.add_tags()

        tags_self = self.dict_meta.copy()
        tags_audio = list(audio.tags)

        for tag_key, value in tags_self.items():

            id3_tag = am.dict_tag2id3.get(tag_key)
            # load existing frame (or not)
            if id3_tag == "TXXX":
                txxx_tags = audio.tags.getall("TXXX")
                for tx in txxx_tags:
                    if tag_key in tx.desc.casefold():
                        frame = [tx]
                        break
                else:
                    frame = []
            else:
                frame = audio.tags.getall(id3_tag)
            if frame:
                if len(frame) > 1:
                    raise Exception(f"to many tags of {id3_tag} found")
                else:
                    frame = frame[0]
                tags_audio.remove(frame.HashKey)
                # if it exists, check if it has to be deleted
                if Empty.is_empty(value) and not write_empty:
                    self._changed_tags.append(("delall", frame.FrameID, value, "*"))
                    audio.tags.delall(frame.FrameID)
                    continue
                elif value is None:
                    if remove_existing:
                        self._changed_tags.append(("delall", frame.FrameID, value, "*"))
                        audio.tags.delall(frame.FrameID)
                    continue
            else:
                if value is None:
                    continue
                # if it does not exist, create a new one and add it to file
                frame = eval("mutagen.id3.{}()".format(id3_tag))
                audio.tags.add(frame)

            if Empty.is_empty(value):
                value = ""

            frame_type = type(frame)
            frame_type_parent = frame_type.mro()[1]

            if frame_type is mutagen.id3.TXXX:
                if tag_key not in frame.desc.casefold():
                    self._set_value(frame, "desc", tag_key)

            if frame.FrameID == "APIC":
                self._fill_apic_frame_with_albumart(frame, value)
            elif frame_type_parent is mutagen.id3.PairedTextFrame:
                # text frame support dropped
                continue
            else:
                if isinstance(value, list):
                    value = join_str_list(value)
                self._set_value(frame, "text", value)
            if not frame.encoding:
                self._set_value(frame, "encoding", mutagen.id3.Encoding.UTF16)
            try:
                lang = frame.lang
                if not lang:
                    self._set_value(frame, "lang", "eng")
            except AttributeError:
                pass

        if remove_existing:
            for tag, value in (
                MP3File(self.file_path).file_read().unprocessed_tag.items()
            ):
                tags_audio.remove(tag)
                audio.tags.delall(tag)
                self._changed_tags.append(("delall", tag, value, "*"))

        # make sure that every tag was checked
        # it is possible a tag is neither in loaded nor in the unprocessed tags list
        # this happened when a mp3 file had the to comment tags COMM and TXXX:COM,
        # where one off them overwrote the other.
        if not len(tags_audio) == 0:
            if remove_existing:
                for tag in tags_audio:
                    audio.tags.delall(tag)
                    self._changed_tags.append(("delall", tag, "N.A.", "*"))
                    logging.info(f"tags mismatch. Tags {tags_audio} were overlooked.")
            else:
                logging.warning(f"tags mismatch. Tags {tags_audio} were overlooked.")

        if len(self._changed_tags) == 0:
            return 0

        if remove_v1:
            v1 = 0  # ID3v1 tags will be removed
        else:  # 1  # ID3v1 tags will be updated  but not added
            v1 = 2  # ID3v1 tags will be created and / or updated

        logging.debug(f"changed tag: {self._changed_tags}")

        if not dry_run:
            audio.save(v1=v1, v2_version=4, v23_sep=None)
            logging.debug(f"File '{self.file_path}' saved.")

        return 1

    def _fill_apic_frame_with_albumart(self, frame, album_art):
        self._set_value(frame, "desc", album_art.desc)
        self._set_value(frame, "type", album_art.ptype)
        self._set_value(frame, "data", album_art.data)
        self._set_value(frame, "mime", album_art.mime)

    def _set_value(self, frame, attr, value):
        org = getattr(frame, attr)
        setattr(frame, attr, value)
        new = getattr(frame, attr)
        if not org == new:
            self._changed_tags.append((frame, attr, org, new))


types = [MP3File]
"""list of all subclasses of AudioFile in this module"""

loader = MP3File
"""class alias for dynamic loading of available AudioFile subclasses"""



#  Copyright (c) 2020 Johannes Nolte
#  SPDX-License-Identifier: GPL-3.0-or-later

from distutils.dir_util import copy_tree
from distutils.file_util import copy_file

import pytest

import mmusicc
import mmusicc.formats
import mmusicc.util
import mmusicc.util.allocationmap
from ._util import *


@pytest.fixture(scope="session")
def dir_orig_data() -> pathlib.Path:
    return pathlib.Path(__file__).parent.joinpath("data")


@pytest.fixture(scope="module")
def allocation_map(request, dir_orig_data):
    if request.fspath.basename == "test_formats.py":
        path_config = "metadata_config.yaml"
    else:
        path_config = dir_orig_data.parent.parent.joinpath("mmusicc/data/config.yaml")
    abs_path = str(dir_orig_data.joinpath(path_config))
    mmusicc.util.init_allocationmap(abs_path, force=True)
    # equals: init_allocationmap(path_config)
    return mmusicc.util.allocationmap


@pytest.fixture(scope="session")
def audio_loaders() -> dict:
    mmusicc.formats.init()
    # trigger skipping of initialisation
    mmusicc.formats.init()
    # equals: init_formats()
    return mmusicc.formats.loaders


@pytest.fixture(scope="session")
def dir_tmp_root(tmp_path_factory):
    return tmp_path_factory.getbasetemp()


@pytest.fixture(scope="session")
def dir_subpackages(tmp_path_factory, dir_orig_data) -> pathlib.Path:
    files = [p.name for p in dir_orig_data.glob("*.*")]
    temp_dir = tmp_path_factory.mktemp("subpackages", numbered=False)
    for file in files:
        file_source = dir_orig_data.joinpath(file)
        copy_file(str(file_source), str(temp_dir))
    return temp_dir


@pytest.fixture(scope="session")
def path_database(dir_subpackages):
    return dir_subpackages.joinpath("database.db")


@pytest.fixture(scope="function")
def temp_database(tmp_path_factory):
    return tmp_path_factory.mktemp("fuu").joinpath("database.db")


@pytest.fixture(scope="function")
def dir_lib_x_flac(tmp_path_factory, dir_orig_data):
    temp_dir = tmp_path_factory.mktemp("lib_x_")
    s_path = dir_orig_data.joinpath("music_lib_x_flac")
    copy_tree(str(s_path), str(temp_dir))
    return temp_dir


@pytest.fixture(scope="session")
def dir_lib_a_flac(tmp_path_factory, dir_orig_data):
    # original flac data do not overwrite
    temp_dir = tmp_path_factory.mktemp("A_flac", numbered=True)
    s_path = dir_orig_data.joinpath("music_lib", "A_flac")
    copy_tree(str(s_path), str(temp_dir))
    save_info = save_files_hash_and_mtime(temp_dir, touch=True)
    yield temp_dir
    # make sure that the original file was not changed
    assert cmp_files_hash_and_time(temp_dir, save_info) == 12


@pytest.fixture(scope="session")
def dir_lib_b_ogg(tmp_path_factory, dir_orig_data):
    # original mp3 data do not overwrite (to be compared with result)
    temp_dir = tmp_path_factory.mktemp("B_ogg", numbered=False)
    s_path = dir_orig_data.joinpath("music_lib", "B_ogg")
    copy_tree(str(s_path), str(temp_dir))
    return temp_dir


@pytest.fixture(scope="session")
def dir_lib_c_ogg(tmp_path_factory, dir_orig_data):
    # lib with missing and wrong metadata
    temp_dir = tmp_path_factory.mktemp("C_ogg", numbered=False)
    s_path = dir_orig_data.joinpath("music_lib", "C_ogg")
    copy_tree(str(s_path), str(temp_dir))
    return temp_dir


@pytest.fixture(scope="function")
def dir_lib_test(tmp_path_factory):
    return tmp_path_factory.mktemp("libt_")



#  Copyright (c) 2020 Johannes Nolte
#  SPDX-License-Identifier: GPL-3.0-or-later

import enum
from collections import defaultdict

from mmusicc.util.util import text_parser_get
from . import allocationmap as am


class MetadataDict(dict):
    """Dictionary containing tags as keys, values can be reset.

    Args:
        init_value (object, optional): initial value of all values,
            defaults to None.
    """

    def __init__(self, init_value=None):
        super().__init__()
        self._init_value = init_value
        for key in am.list_tags:
            self[key] = init_value

    def import_builtin_dict(self, b_dict) -> dict:
        """Import a dictionary into the MetadataDict, mapping TAGs.

        Args:
            b_dict (dict): any dictionary with tag, value pairs

        Returns:
            list: key value pairs that could't be imported.
        """
        rest = scan_dictionary(b_dict, self)
        return rest

    def reset(self):
        """Reset all values to initial values, defines at instance creation."""
        for key in list(self):
            self[key] = self._init_value

    def copy_not_none(self):
        """Returns a copy of the dictionary omitting items with value None.

        Returns:
            dict: copy of dict items which value is not None.
        """
        return {k: v for k, v in self.items() if v is not None}

    def copy(self):
        return {k: v for k, v in self.items()}

    def convert_none2empty(self):
        """can be used to remove/add all tags that have no value"""
        raise NotImplementedError()


def metadatadict(b_dict) -> MetadataDict:
    """Returns a MetadataDict from the dict, unknown tags are dropped"""
    m_dict = MetadataDict()
    m_dict.import_builtin_dict(b_dict)
    return m_dict


class PairedText(object):
    """Proxy object for mutagen.id3.PairedTextFrames. Experimental not used."""

    def __init__(self, role, name):
        if role == "":
            role = None
        self.role = role
        self.name = name

    def __repr__(self):
        if not self.role:
            return self.name
        else:
            return "{}: {}".format(self.role, self.name)

    def __iter__(self):
        return [self.role, self.name]


class Empty(object):
    """Object representing the emptiness of a existing tag without an value."""

    value = ""

    def __repr__(self):
        return "<empty>"

    def __eq__(self, other):
        if isinstance(other, Empty):
            return True
        else:
            return False

    def __hash__(self):
        return 42

    @staticmethod
    def is_empty(text) -> bool:
        """Returns True if text is instance Empty or a empty string ''."""
        if isinstance(text, Empty):
            return True
        if isinstance(text, str) and text.strip() == "":
            return True
        return False

    @staticmethod
    def is_empty_or_none(text) -> bool:
        """Returns True if text is instance Empty, a empty string or None."""
        if text is None or Empty.is_empty(text):
            return True
        return False


class AlbumArt(object):
    def __init__(self):
        self.ptype = 0
        self.mime = u""
        self.desc = u""  # filename
        self.data = b""

    def __eq__(self, other):
        if isinstance(other, AlbumArt):
            return self.data_hash == other.data_hash and self.ptype == other.ptype
        else:
            return False

    def __repr__(self):
        return self.desc

    def __hash__(self):
        return hash((self.data_hash, self.ptype, self.desc))

    # might be useful in ne feature
    # def size(self):
    #     from PIL import Image
    #     import io
    #     img = Image.open(io.BytesIO(self.data))
    #     return img.size

    @property
    def data_hash(self):
        # not good but better than nothing
        return len(self.data)


class PictureType(enum.Enum):
    """Enumeration of image types defined by the ID3 standard for the APIC
    frame, but also reused in WMA/FLAC/VorbisComment.
    """

    OTHER = 0
    """Other"""

    FILE_ICON = 1
    """32x32 pixels 'file icon' (PNG only)"""

    OTHER_FILE_ICON = 2
    """Other file icon"""

    COVER_FRONT = 3
    """Cover (front)"""

    COVER_BACK = 4
    """Cover (back)"""

    LEAFLET_PAGE = 5
    """Leaflet page"""

    MEDIA = 6
    """Media (e.g. label side of CD)"""

    LEAD_ARTIST = 7
    """Lead artist/lead performer/soloist"""

    ARTIST = 8
    """Artist/performer"""

    CONDUCTOR = 9
    """Conductor"""

    BAND = 10
    """Band/Orchestra"""

    COMPOSER = 11
    """Composer"""

    LYRICIST = 12
    """Lyricist/text writer"""

    RECORDING_LOCATION = 13
    """Recording Location"""

    DURING_RECORDING = 14
    """During recording"""

    DURING_PERFORMANCE = 15
    """During performance"""

    SCREEN_CAPTURE = 16
    """Movie/video screen capture"""

    FISH = 17
    """A bright coloured fish"""

    ILLUSTRATION = 18
    """Illustration"""

    BAND_LOGOTYPE = 19
    """Band/artist logotype"""

    PUBLISHER_LOGOTYPE = 20
    """Publisher/Studio logotype"""


class Div(object):
    """
    Object Representing a group different values. Provides rich comparison.

    Still Under Construction.

    Args:
        key                      (str, optional): Dictionary key which values
            to represent
        list_metadata (list<Metadata>, optional): List of Metadata object
            containing the values.
    """

    def __init__(self, key=None, list_metadata=None):
        self._dict_values = dict()
        self._diff = None
        self._key_tag = None  # maybe not needed
        if key and list_metadata:
            self.add_metadata(key, list_metadata)

    def __repr__(self):
        return "<div>"

    def __eq__(self, other):
        if not isinstance(other, Div):
            return False
        equal = None
        for meta_a in list(self._dict_values):
            for meta_b in list(other._dict_values):
                if meta_a.file_name == meta_b.file_name:
                    if not (
                        self._dict_values.get(meta_a) == other._dict_values.get(meta_b)
                    ):
                        equal = False
        if equal is None:
            equal = False
        return equal

    def add_metadata(self, key_tag, list_metadata):
        """Adds the values of a key of a metadata objects to div list.

        Args:
            key_tag (str): Key of values to be added.
            list_metadata (list of Metadata): List of metadata which values to
                be added.
        """
        self._key_tag = key_tag
        for metadata in list_metadata:
            self.add_value(metadata, metadata.get_tag(self._key_tag))

    def add_value(self, metadata, value):
        """Adds one value to the Div object.
        
        Args:
            metadata: Metadata object the value belongs too.
            value: Value that is different.
        """
        self._dict_values[metadata] = value


def scan_dictionary(dict_tags, dict_data, parse_text=False):
    """Scan a dictionary (dict_tags) for tags and fill dict_data with them.

    Args:
        dict_tags (dict): dictionary (tag_str: value) items to be imported.
        dict_data   (dict): dictionary to be filled with imported tag values.
        parse_text  (bool): if True, pass values to text_parser_get else return value
            as it is. Defaults to False.
    Returns:
        dict<str, str>: dictionary with all tags whose name could not be
            associated with. Note: keys are casfolded.
    """

    # Make a copy of the source dictionary, so it is unchanged
    dict_dummy = dict_tags.copy()
    dict_dummy = {k.casefold(): v for k, v in dict_dummy.items()}

    # filled with list so the same tag can have multiple values
    dict_tmp = defaultdict(list)

    # find a tag for each key in scanned dictionary
    for key_str in list(dict_dummy):
        try:
            tag_key = am.dict_str2tag[key_str]
            tag_val = dict_dummy.pop(key_str)
        except KeyError:
            continue

        if Empty.is_empty(tag_val):
            tag_val = Empty()

        if parse_text:
            dict_tmp[tag_key].extend(text_parser_get(tag_val))
        else:
            dict_tmp[tag_key].append(tag_val)

    for key, value in dict_tmp.items():
        if len(value) == 1:
            dict_data[key] = value[0]
        else:
            dict_data[key] = value

    return dict_dummy
